#ifndef PLEDGE_LIBC_STR_STR_H_
#define PLEDGE_LIBC_STR_STR_H_

#include "libc/integral/c.h"

#include <ctype.h>
#include <stdbool.h>
#include <string.h>
#include <strings.h>

bool endswith(const char *, const char *) strlenesque;

#endif
